#include <stdio.h>

#include "make_keys.hpp"
#include "common.hpp"
#include "gen_random.hpp"
#include "seek_bigger_prime.hpp"
#include "countbits.hpp"
#include "RSA-core.hpp"
#include "numberio.hpp"

#include <stdlib.h>

void create_p (RSA &rsa, int bits)
{
  printf("*** Creating p\n");
  gen_random (bits, rsa.p);
  seek_bigger_prime (rsa.p);
};

void create_q (RSA &rsa, int bits)
{
  printf("*** Creating q\n");
  gen_random (bits, rsa.q);
  seek_bigger_prime (rsa.q);
};

void calc_n (RSA &rsa)
{
  printf("*** Calculating n\n");
  rsa.calc_n_from_p_q ();
};

void calc_phin (RSA &rsa)
{
  printf("*** Calculating phi(n)\n");
  rsa.calc_phin_from_p_q ();
};

void create_s (RSA &rsa)
{
  printf("*** Creating s\n");

  mpz_t x;
  mpz_init (x);

  // 2^s >= n  is a security criteria
  mpz_set_ui (rsa.s, countbits(rsa.n) + 1);

  // i am paranoid. give it some extra size
  mpz_add_ui (rsa.s, rsa.s, 100000000);

  // now increase s until it matches RSA criteria
 again:
  mpz_gcd (x, rsa.s, rsa.phin);
  if (mpz_cmp_ui (x, 1) != 0) {
    mpz_add_ui (rsa.s, rsa.s, 1);
    goto again;
  };

  if (mpz_cmp (rsa.s, rsa.phin) >= 0) {
    fprintf(stderr,"s >= phi(n)\n");
    exit(1);
  };

  if (mpz_cmp_ui (rsa.s, 0) <= 0) {
    fprintf(stderr,"s <= 0\n");
    exit(1);
  };

  mpz_clear (x);
};

void calc_e (RSA &rsa)
{
  printf ("*** Calculating e...\n");
  rsa.calc_e_from_s_phin ();

  if (mpz_cmp (rsa.e, rsa.phin) >= 0) {
    fprintf(stderr,"e >= phi(n)\n");
    exit(1);
  };

  if (mpz_cmp_ui (rsa.e, 0) <= 0) {
    fprintf(stderr,"e <= 0\n");
    exit(1);
  };

  if (countbits(rsa.e) <= countbits(rsa.n)/2) {
    fprintf(stderr,"security problem: e is up to one quarter of the size of n\n");
    fprintf(stderr,"this is very unlikly to happen. please re-run key generation\n");
    exit(1);
  };
};

int make_keys (int argc, char *argv[])
{
  if (argc!=5) usage ();

  RSA rsa;

  mpz_t x;
  mpz_init (x);

  int bits = atol (argv[2]);

  if (bits<128) {
    printf("Keys with less than 128 bits are not allowed.\n");
    printf("You wanted at least a bit of security, didn't you ?\n");
    exit(1);
  };

  printf("i will now generate two random numbers using Linux /dev/random\n"
	 "entropy source. you can speed up this process by moving\n"
	 "your mouse and or beating on your keys and or creating\n"
	 "harddisk activity and or many other things\n"
	 "see kernel source if you are curious\n\n");

  create_p (rsa, bits >> 1);
  //  show ("p = ", rsa.p);
 again_create_q:
  create_q (rsa, bits >> 1);
  if (mpz_cmp(rsa.p,rsa.q)==0) {
    printf("Oh, p == q ! I create a new q\n");
    goto again_create_q;
  };
  //  show ("q = ", rsa.q);
  calc_n (rsa);
  //  show ("n = ", rsa.n);
  calc_phin (rsa);
  //  show ("phi(n) = ", rsa.phin);
  create_s (rsa);
  //  show ("s = ", rsa.s);
  calc_e (rsa);
  //  show ("e = ", rsa.e);

  FILE *f;
  assert ((f = fopen (argv[3], "w"))!=0);
  fprintf (f, "yyyRSA SECRET KEY V2\n");
  number_write (f, rsa.n);
  number_write (f, rsa.e);
  assert (ferror(f)==0);
  assert (fclose(f)==0);

  assert ((f = fopen (argv[4], "w"))!=0);
  fprintf (f, "yyyRSA PUBLIC KEY V2\n");
  number_write (f, rsa.n);
  number_write (f, rsa.s);
  assert (ferror(f)==0);
  assert (fclose(f)==0);

  printf (
	  "The public and secret files are now generated.\n"
	  "If you want to hack the keys, i.e create the secret\n"
	  "key using the public key, you have to factorize\n"
	  "a number that consists of two prime numbers\n"
	  "and has a size of %d bits.\n"
	  , countbits(rsa.n)
	  );
  printf ("If you want to see this number, here it is:\n");
  mpz_out_str (stdout, 10, rsa.n);
  printf ("\n\n");
  
  mpz_clear (x);

  return 0;
};
